#ifndef OOPSMP_MOTION_PLANNERS_PRM_VISPRM__OOPSMP_VISPRM_H_
#define OOPSMP_MOTION_PLANNERS_PRM_VISPRM__OOPSMP_VISPRM_H_

#include "MotionPlanners/PRM/OOPSMPPRM.H"

namespace OOPSMP
{
    ForwardClassDeclaration(VisPRM);

    class VisPRM : public PRM
    {
    protected:
	class GuardVisibility
	{
	    friend class VisPRM;

	protected:
	    int nr_connections;   // the number of roadmap nodes from different connected components
	                          // that the current sample is visible from
	    int* connections;     // the indices of the nodes from different connected components that can
	                          // view the state
	    Path_t* paths;        // the paths that connect the state to roadmap nodes which can "see" the state
	    
	    int allocated;        // number of allocated connections and paths	    

	public:
	    GuardVisibility()
	    {
		nr_connections = 0; 
		allocated = 10; 
		connections = (int*)malloc(allocated*sizeof(int));
		paths = (Path_t*)malloc(allocated*sizeof(Path_t));
	    }

	    ~GuardVisibility()
	    {
		free(connections);
		free(paths);
	    }

	    void clearGuardVisibility()
	    {
		if( nr_connections == 1 )
		    delete( paths[0] );
		nr_connections = 0;
	    }

	    void addConnection( int index, Path_t path )
	    {
		nr_connections++;
		if( nr_connections >= allocated )
		{
		    allocated *= 2;
		    connections = (int*)realloc(connections, allocated*sizeof(int));
		    paths = (Path_t*)realloc(paths, allocated*sizeof(Path_t));
		}	

		connections[nr_connections-1] = index;
		paths[nr_connections-1] = path;
	    }

	    // is the current state a guard
	    bool isGuard() { return (nr_connections == 0); }
	    
	    // is the current state a connector
	    bool isConnector() { return (nr_connections >= 2); }	

	};
	typedef GuardVisibility* GuardVisibility_t;


	int max_tries;        //maximum failed attempts to create a new guard

	State_t  state;	      // the current sample which is a candidate to be added to the roadmap

	GuardVisibility_t g_vis; // the guard visibility data structure recomputed for each sample

	StateSpace_t        ss;
	ValidStateSampler_t vss;
	LocalPlanner_t      lp;

    public:
	VisPRM();
	virtual ~VisPRM();
	
	virtual void clear();
	virtual void preprocess( const double ptime );

    protected:
	// is the sampled state already visible from a roadmap on the same
	// component as the the node: "node_index"?
	bool isNodeVisibleFromComponent( int node_index);

	// computes the guards from different connected components that can view the state
	// limited_vis: whether we impose a limit on the visibility of nodes
	// d_vis:       the distance limit for visibility
	virtual void computeGuardVisibility( );
	
	// sample a random valid state
	void sampleState( );

	// add the sampled state as a guard
	void addGuard( );
	
    };
    typedef VisPRM* VisPRM_t;

    DeclareInstanceFactory(VisPRM, PRMFactory);
}

#endif
